﻿using System;
using System.Diagnostics;
using System.Globalization;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA
{
    public partial class EasyUAClientTest
    {
        #region Static Private Methods

        static private int Test_Subscribe_Unsubscribe(
            [NotNull] IEasyUAClient client, 
            [NotNull] UAEndpointDescriptor endpointDescriptor, 
            [NotNull] UANodeDescriptor nodeDescriptor, 
            int duration)
        {
            TestSection.Arrange();

            int eventCount = 0;
            client.DataChangeNotification += (sender, args) =>
            {
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Event #{0} at {1}:", eventCount, DateTime.Now));
                eventCount++;
                // ReSharper disable once ConvertToLambdaExpression
                AssertUtilities.WriteLine(args);
                // IMPROVE: do more
            };

            TestSection.Act();

            AssertUtilities.Phase("Subscribing");
            int handle = client.SubscribeDataChange(endpointDescriptor, nodeDescriptor, 1000, state:"TestState");

            AssertUtilities.WaitingPhase(duration);

            AssertUtilities.Phase("Unsubscribing");
            client.UnsubscribeMonitoredItem(handle);

            return eventCount;
        }

        static private int Test_Subscribe_Unsubscribe_Callback(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor,
            [NotNull] UANodeDescriptor nodeDescriptor,
            int duration)
        {
            TestSection.Arrange();

            int eventCount = 0;
            //client.DataChangeNotification += (sender, args) =>
            //{
            //    AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Event #{0} at {1}:", eventCount, DateTime.Now));
            //    eventCount++;
            //    // ReSharper disable once ConvertToLambdaExpression
            //    AssertUtilities.WriteLine(args);
            //    // IMPROVE: do more
            //};

            TestSection.Act();

            AssertUtilities.Phase("Subscribing");
            int handle = client.SubscribeDataChange(
                endpointDescriptor: endpointDescriptor,
                nodeDescriptor: nodeDescriptor,
                samplingInterval: 1000,
                dataChangeCallback: (sender, args) =>
                {
                    AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Event #{0} at {1}:",
                        eventCount, DateTime.Now));
                    eventCount++;
                    // ReSharper disable once ConvertToLambdaExpression
                    AssertUtilities.WriteLine(args);
                    // IMPROVE: do more
                },
                state: "TestState");

            AssertUtilities.WaitingPhase(duration);

            AssertUtilities.Phase("Unsubscribing");
            client.UnsubscribeMonitoredItem(handle);

            return eventCount;
        }

        #endregion



        #region Test Methods
        // ReSharper disable InconsistentNaming

        [Timeout(2 * 60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void Test_Subscribe_Change_Unsubscribe()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            //UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:4842";
            //UAEndpointDescriptor endpointDescriptor = TestDefs.EndpointUrlStrings.SampleServer_Http;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            AssertUtilities.Phase("Initializing...");
            var target = new EasyUAClient();
            int eventCount = 0;
            target.DataChangeNotification += (sender, args) =>
            {
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Event #{0} at {1}:", eventCount, DateTime.Now));
                eventCount++;
                // ReSharper disable once ConvertToLambdaExpression
                AssertUtilities.WriteLine(args);
                // IMPROVE: do more
            };

            TestSection.Act();

            AssertUtilities.Phase("Subscribing");
            int handle = target.SubscribeDataChange(
                endpointDescriptor,
                OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue,
                1000);

            AssertUtilities.WaitingPhase(10 * 1000);

            AssertUtilities.Phase("Changing subscription (1)");
            target.ChangeMonitoredItemSubscription(handle, 1500);

            AssertUtilities.WaitingPhase(10 * 1000);

            AssertUtilities.Phase("Changing subscription (2)");
            target.ChangeMonitoredItemSubscription(handle, 2200);

            AssertUtilities.WaitingPhase(10 * 1000);

            AssertUtilities.Phase("Unsubscribing");
            target.UnsubscribeMonitoredItem(handle);

            AssertUtilities.WaitingPhase(5 * 1000);

            TestSection.Assert();

            AssertUtilities.Phase("Evaluating");
            AssertUtilities.WriteLine($"EventCount: {eventCount}");

            // IMPROVE: we used to have an error where ModifyMonitoredItems call was made (which it shouldn't). Test it.

            AssertUtilities.Done();
        }

        [Timeout(2 * 60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void Test_Subscribe_Unsubscribe_Dispose()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            //UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:4842";
            //UAEndpointDescriptor endpointDescriptor = TestDefs.EndpointUrlStrings.SampleServer_Http;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            AssertUtilities.Phase("Initializing");
            using (var target = new EasyUAClient { Isolated = true })
            {
                UANodeDescriptor nodeDescriptor = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;

                // ReSharper disable once RedundantArgumentName
                int eventCount = Test_Subscribe_Unsubscribe(target, endpointDescriptor, nodeDescriptor, duration: 60 * 1000);

                TestSection.Assert();

                AssertUtilities.Phase("Finalizing");
                AssertUtilities.WriteLine(eventCount);
            }

            AssertUtilities.Done();
        }

        [Timeout(2 * 60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void Test_Subscribe_Unsubscribe_Dispose_KepServerEx()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = Kepware_KepServerEx.EndpointUrlString_OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            using (var target = new EasyUAClient { Isolated = true })
            {
                UANodeDescriptor nodeDescriptor = "nsu=KEPServerEX;ns=2;s=_System._Time_Second";

                // ReSharper disable once RedundantArgumentName
                int eventCount = Test_Subscribe_Unsubscribe(target, endpointDescriptor, nodeDescriptor, duration: 60 * 1000);

                TestSection.Assert();

                AssertUtilities.Phase("Finalizing");
                AssertUtilities.WriteLine(eventCount);
            }
        }

        [Timeout(2 * 60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void Test_Subscribe_Unsubscribe_Dispose_Wait_Long()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            //UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:4842";
            //UAEndpointDescriptor endpointDescriptor = TestDefs.EndpointUrlStrings.SampleServer_Http;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            const int millisecondsWait = 8 * 1000;

            using (var target = new EasyUAClient { Isolated = true })
            {
                UANodeDescriptor nodeDescriptor = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;

                // ReSharper disable once RedundantArgumentName
                int eventCount = Test_Subscribe_Unsubscribe(target, endpointDescriptor, nodeDescriptor, duration: 10 * 1000);

                TestSection.Assert();

                AssertUtilities.Phase("Finalizing");
                AssertUtilities.WriteLine(eventCount);
            }

            AssertUtilities.WaitingPhase(millisecondsWait);

            // IMPROVE: test that we have truly called CloseSession
        }

        [Timeout(2 * 60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void Test_Subscribe_Unsubscribe()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            //UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:4842";
            //UAEndpointDescriptor endpointDescriptor = TestDefs.EndpointUrlStrings.SampleServer_Http;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UANodeDescriptor nodeDescriptor = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;

            var target = new EasyUAClient { Isolated = true };
            // ReSharper disable once RedundantArgumentName
            int eventCount = Test_Subscribe_Unsubscribe(target, endpointDescriptor, nodeDescriptor, duration: 10 * 1000);

            TestSection.Assert();

            AssertUtilities.Phase("Finalizing");
            AssertUtilities.WriteLine(eventCount);

            // IMPROVE: test that we have truly called CloseSession
        }

        [Timeout(2 * 60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void Test_Subscribe_Unsubscribe_Callback()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            //UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:4842";
            //UAEndpointDescriptor endpointDescriptor = TestDefs.EndpointUrlStrings.SampleServer_Http;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UANodeDescriptor nodeDescriptor = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;

            var target = new EasyUAClient { Isolated = true };
            // ReSharper disable once RedundantArgumentName
            int eventCount = Test_Subscribe_Unsubscribe_Callback(target, endpointDescriptor, nodeDescriptor, duration: 10 * 1000);

            TestSection.Assert();

            AssertUtilities.Phase("Finalizing");
            AssertUtilities.WriteLine(eventCount);
        }

        [Timeout(2 * 60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void Test_Subscribe_Unsubscribe_1min()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            //UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:4842";
            //UAEndpointDescriptor endpointDescriptor = TestDefs.EndpointUrlStrings.SampleServer_Http;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UANodeDescriptor nodeDescriptor = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;

            var target = new EasyUAClient { Isolated = true };
            // ReSharper disable once RedundantArgumentName
            int eventCount = Test_Subscribe_Unsubscribe(target, endpointDescriptor, nodeDescriptor, duration: 60 * 1000);

            TestSection.Assert();

            AssertUtilities.Phase("Finalizing");
            AssertUtilities.WriteLine(eventCount);

            // IMPROVE: test that we have truly called CloseSession
        }

        [Timeout(2 * 60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void Test_Subscribe_Unsubscribe_Wait_Long()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.OpcTcp;
            //UAEndpointDescriptor endpointDescriptor = "opc.tcp://localhost:4842";
            //UAEndpointDescriptor endpointDescriptor = TestDefs.EndpointUrlStrings.SampleServer_Http;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            const int millisecondsWait = 8 * 1000;

            UANodeDescriptor nodeDescriptor = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;

            var target = new EasyUAClient { Isolated = true };
            Debug.Assert(millisecondsWait > target.IsolatedParameters.SessionParameters.HoldPeriod + 2 * 1000);
            // ReSharper disable once RedundantArgumentName
            int eventCount = Test_Subscribe_Unsubscribe(target, endpointDescriptor, nodeDescriptor, duration: 10 * 1000);

            AssertUtilities.WaitingPhase(millisecondsWait);

            TestSection.Assert();

            AssertUtilities.Phase("Finalizing");
            AssertUtilities.WriteLine(eventCount);

            // IMPROVE: test that we have truly called CloseSession
        }

        // ReSharper restore InconsistentNaming
        #endregion



        #region Private Methods

        #endregion
    }
}
